/*=======================================================================*
 |   file name : _gcursor.e
 |-----------------------------------------------------------------------*
 |   function  : emulate emacs brace matching
 |-----------------------------------------------------------------------*
 |   author    : Gregg Tavares    http://greggman.com
 *=======================================================================*/

/*
    To use these macros you need to bind the following keys
    
    Up        -> gat_go_up
    Down      -> gat_go_down
    Left      -> gat_go_left
    Right     -> gat_go_right
    Backspace -> gat_backspace
    End       -> gat_enter_endofline
    )         -> gat_enter_endparan
    ]         -> gat_enter_endsqaurebracked
    }         -> gat_enter_endbrace
                   note: you need to bind it once with
                         Affect All modes checked and again
                         with it not checked while you are editing
                         a C or CPP file
    
    This will make it *mostly* like emacs brace matching.
    
    There are a few issues.
    
    *) It appears there is no way to have an *idle* macro.  Meaning that
       in emacs, no matter how the cursor gets somewhere, if it's positioned
       past a ) or ] or } the corresponding character will be highlighted.
       
       I couldn't figure out how to do that in slickedit so I had to replace
       the key defs above.  That means there are cases were it will not highlight.
       For example if you paste a ) in.
       
    *) The mouse will get messed up.  The problem is the only way to highlight
       was to do a selection.  So, I find the matching brace, make a one character
       selection then I busy wait for another key.  That busy waiting messes
       up the mouse.  As an example put the cursor after a ) so that you see
       the corresponding ( highlighted.  Try to use the menus with the mouse.
       :-(
       
       It's not a problem for me because I pretty much never use the mouse
       inside slickedit.
       
    *) Back to the idle issue, I don't know how to make it generic so that
       it would just work with however you have slickedit configured so
       overridding some of those keys will make it work like I have my
       editor configured.  If you don't like it you'll have to modify
       the macros.  For example gat_backspace is hard coded to do a 
       linewrap_rubout and gat-enter-endofline works like it did in Brief;
       one press is the end of line, two is the end of the screen, 3 is the
       end of the buffer.
       
    *) One advantage over emacs IMO is that I prefer this style of C
       
            if (condition)
            {
               code
            }
            
       An emacs friend of mine pointed out that in emacs if you use this style

            if (condition) {
               code
            }
            
       since emacs shows you the matching line in the status area you get
       to see the entire line as you type the closing brace.  Well, I fixed
       this macro so that if when you type a closing brace if the matching
       opening brace is on a line all by itsself the macro will show
       the previous line with " {" appended to it.
       
       It would be nice if there was a way to format the messages though.
       In emacs the line displayed in the status area has the paren/brace
       highlighted.  In slickedit there is no highlighting in the
       status bar :-(  So for example if the opening like is ((( and it's
       off the screen you can't tell which paren you are matching by looking
       at the status bar like you can in emacs.  I tried surrouding the
       matching character with stuff like
       
       (-=>(<=-)
       
       but as you can see it just really doesn't work.  It really needs
       to be in color.
       
    <wish>
    
    I really just wish these macros were not necessary and that emacs like
    brace,paren,bracket matching was just another option in slickedit including
    showing the matching line in the status bar with the matching character
    highlighted in color.
    
    Slickedit is great and they claim to emulate emacs to some extent.  You'd
    think this would be one of those important ways of emulation.
    
    If they won't add emacs style brace matching I wis they would at least
    fix/add whatever is needed to make these macros work better.  That includes
    
      *) A way to make it work and not mess up the mouse
    
      *) some kind of idle system so I could just check after every key press
         instead of binding to all the keys above
         
      *) a way to allow color in the status bar.
      
    </wish>
    
*/

#include 'slick.sh'

void gat_blink_matching()
{
   if (!command_state() && !_macro ('R') && !p_hex_mode)
   {
      if (_get_selinfo (junk_col, junk_ecol, junk_buf_id) == TEXT_NOT_SELECTED_RC)
      {
         get_line line;

         line    = expand_tabs(line);
         cOn     = substr (line, p_col, 1);
         if (p_col > 1)
         {
            cBefore = substr (line, p_col - 1, 1);
         }
         else
         {
            cBefore = '';
         }

         if (pos(cBefore, ')]}') || pos(cOn, '({['))
         {
            typeless p;
            typeless p2;

            _save_pos2(p);
            boolean restored=false;

            if (pos(cBefore, ')]}'))
            {
               left ();
            }

            if (find_matching_paren() == 0)
            {
               _str oldsel;

               get_line line;
               line = expand_tabs(line);
               // check if line is only "{" and if so get the previous line
               if (strip(line) :== "{")
               {
                  up();
                  get_line line;
                  line = line :+ " {";
                  down();
               }

               if (!save_selection (oldsel))
               {
                  mark_id = _alloc_selection();
                  if (mark_id >= 0)
                  {
                     _select_char(mark_id);
                     right();
                     _select_char(mark_id);
                     _show_selection(mark_id);
                     _restore_pos2(p);
                     restored=true;

                     message (line);
                     
                     refresh();
                     mou_capture()
                     while (test_event('R') :== "")
                     {
                        // just so it's not a complete busy wait
                        delay(30,'K');
                     }
                     mou_release();
                     
                     _free_selection(mark_id);
                  }
                  restore_selection (oldsel);
               }
            }
            if(!restored) {
               _restore_pos2(p);
            }
         }
      }
   }
}

static int updown_col=0;
static void gat_stay_on_text(updown_col)
{
   if ( updown_col && (! select_active() || _select_type():!='BLOCK') ) {
      if ( _text_colc(0,'E')<updown_col ) {
         p_col=_text_colc(0,'E')+1;
      } else if ( _text_colc(updown_col,'T')<0 ) {
         p_col=_text_colc(1-_text_colc(updown_col,'T'),'i');
      } else {
         p_col=updown_col;
      }
   }

}
_command gat_go_up() name_info(','VSARG2_READ_ONLY|VSARG2_TEXT_BOX|VSARG2_CMDLINE|VSARG2_EDITORCTL|VSARG2_REQUIRES_EDITORCTL|VSARG2_LINEHEX)
{
   if (command_state() || p_hex_mode)
   {
      _cursor_updown(arg(1),"",1);
   }
   else
   {
      prev_cmd=translate(name_name(prev_index('','C')),'_','-');
      if (prev_cmd!='find_matching_paren' && prev_cmd!='gat_go_up' && prev_cmd!='gat_go_down' && def_updown_col) {
         updown_col=p_col;
      }
      _cursor_updown(arg(1),"",1);
      if (def_updown_col) {
         gat_stay_on_text(updown_col);
      }
      gat_blink_matching();
   }
}

_command gat_go_down() name_info(','VSARG2_READ_ONLY|VSARG2_TEXT_BOX|VSARG2_CMDLINE|VSARG2_EDITORCTL|VSARG2_REQUIRES_EDITORCTL|VSARG2_LINEHEX)
{
   if (command_state() || p_hex_mode)
   {
      _cursor_updown(arg(1),"n",1);
   }
   else
   {
      prev_cmd=translate(name_name(prev_index('','C')),'_','-');
      if (prev_cmd!='find_matching_paren' && prev_cmd!='gat_go_up' && prev_cmd!='gat_go_down' && def_updown_col) {
         updown_col=p_col;
      }
      _cursor_updown(arg(1),"n",1);
      if (def_updown_col) {
         gat_stay_on_text(updown_col);
      }
      gat_blink_matching();
   }
}

_command gat_go_left() name_info(','VSARG2_READ_ONLY|VSARG2_TEXT_BOX|VSARG2_REQUIRES_EDITORCTL|VSARG2_LINEHEX)
{
   cursor_left();
   if (!command_state())
   {
      updown_col=p_col;
   }
   gat_blink_matching();
}
_command gat_go_right() name_info(','VSARG2_READ_ONLY|VSARG2_TEXT_BOX|VSARG2_REQUIRES_EDITORCTL|VSARG2_LINEHEX)
{
   cursor_right();
   if (!command_state())
   {
      updown_col=p_col;
   }
   gat_blink_matching();
}
_command gat_enter_endbrace() name_info(','MARK_ARG2|TEXT_BOX_ARG2|EDITORCTL_ARG2)
{
   keyin("}");
   gat_blink_matching();
}
_command gat_enter_endsquarebracket() name_info(','MARK_ARG2|TEXT_BOX_ARG2|EDITORCTL_ARG2)
{
   keyin("]");
   gat_blink_matching();
}
_command gat_enter_endparen() name_info(','MARK_ARG2|TEXT_BOX_ARG2|EDITORCTL_ARG2)
{
   keyin(")");
   gat_blink_matching();
}

_command gat_backspace() name_info(','VSARG2_READ_ONLY|VSARG2_TEXT_BOX|VSARG2_CMDLINE|VSARG2_EDITORCTL|VSARG2_REQUIRES_EDITORCTL|VSARG2_LINEHEX)
{
   linewrap_rubout();
   gat_blink_matching();
}

// gat_enter_endofline should be bound to END
static _str homeend_count=0;
_command gat_enter_endofline() name_info(','VSARG2_READ_ONLY|VSARG2_TEXT_BOX|VSARG2_CMDLINE|VSARG2_EDITORCTL|VSARG2_REQUIRES_EDITORCTL|VSARG2_LINEHEX)
{
   if ( command_state() || p_hex_mode ) {
      return(end_line())
   }
   index=last_index('','C')
   first=name_name(prev_index('','C'))!='gat-enter-endofline'
   if (first || homeend_count>=3) {
      homeend_count=0;
   }
   ++homeend_count;
   _macro_delete_line()
   switch (homeend_count) {
   case 1:
      _macro_call('end_line')
      end_line();
      break;
   case 2:
      _macro_append('bottom_of_window();_end_line();');
      bottom_of_window();_end_line();
      break;
   case 3:
      _macro_call('bottom_of_buffer')
      bottom_of_buffer();
      break;
   }
   gat_blink_matching();
   last_index(index,'C')

}


